from typing import Any, Mapping


class ConfigurationError(Exception):
    invalid_data: Mapping[str, Any]

    def __init__(self, invalid_data: Mapping[str, Any]) -> None:
        super().__init__(invalid_data)
        self.invalid_data = invalid_data


class UnknownImageReference(ValueError):
    """
    Represents an error for invalid/unknown image reference.
    The first argument of this exception should be the reference given by the user.
    """

    def __str__(self) -> str:
        return f"Unknown image reference: {self.args[0]}"


class ImageNotAvailable(ValueError):
    """
    Represents an error for unavailability of the image in agents.
    The first argument of this exception should be the reference given by the user.
    """

    def __str__(self) -> str:
        return f"Unavailable image in the agent: {self.args[0]}"


class UnknownImageRegistry(ValueError):
    """
    Represents an error for invalid/unknown image registry.
    The first argument of this exception should be the registry given by the user.
    """

    def __str__(self) -> str:
        return f"Unknown image registry: {self.args[0]}"


class InvalidImageName(ValueError):
    """
    Represents an invalid string for image name.
    """

    def __str__(self) -> str:
        return f"Invalid image name: {self.args[0]}"


class InvalidImageTag(ValueError):
    """
    Represents an invalid string for image tag and full image name.
    Image tag should be a string of form below

    ```
    <version-stringA>-<platform-tag-1A>-<platform-tag-2A>-....
    ```
    """

    def __init__(self, tag: str, full_name: str | None = None) -> None:
        super().__init__(tag, full_name)
        self._tag = tag
        self._full_name = full_name

    def __str__(self) -> str:
        if self._full_name is not None:
            return f"Invalid or duplicate image name tag: {self._tag}, full image name: {self._full_name}"
        else:
            return f"Invalid or duplicate image name tag: {self._tag}"


class AliasResolutionFailed(ValueError):
    """
    Represents an alias resolution failure.
    The first argument of this exception should be the alias given by the user.
    """

    def __str__(self) -> str:
        return f"Failed to resolve alias: {self.args[0]}"


class InvalidIpAddressValue(ValueError):
    """
    Represents an invalid value for ip_address.
    """


class VolumeMountFailed(RuntimeError):
    """
    Represents a mount process failure.
    """


class VolumeUnmountFailed(RuntimeError):
    """
    Represents a umount process failure.
    """



from typing import Annotated, Mapping, Sequence, TypeAlias

from lark import Lark, Transformer, lexer
from lark.exceptions import LarkError

_grammar = r"""
    start: pair ("," pair)*
    pair: key [("="|":") value]
    key: SLASH? CNAME (SEPARATOR|CNAME|DIGIT)*
    value: SLASH? CNAME (SEPARATOR|CNAME|DIGIT)* | ESCAPED_STRING

    SEPARATOR: SLASH | "\\," | "\\=" | "\\:" | DASH
    SLASH: "/"
    DASH: "-"

    %import common.CNAME
    %import common.DIGIT
    %import common.ESCAPED_STRING
    %import common.WS
    %ignore WS
"""

PairType: TypeAlias = tuple[str, str]


class DictTransformer(Transformer):
    reserved_keys = frozenset({"type", "source", "target", "perm", "permission"})

    def start(self, pairs: Sequence[PairType]) -> Mapping[str, str]:
        if pairs[0][0] not in self.reserved_keys:  # [["vf-000", "/home/work"]]
            result = {"source": pairs[0][0]}
            if target := pairs[0][1]:
                result["target"] = target
            return result
        return dict(pairs)  # [("type", "bind"), ("source", "vf-000"), ...]

    def pair(self, token: Annotated[Sequence[str], 2]) -> PairType:
        return (token[0], token[1])

    def key(self, token: list[lexer.Token]) -> str:
        return "".join(token)

    def value(self, token: list[lexer.Token]) -> str:
        return "".join(token)


_parser = Lark(_grammar, parser="lalr")


class MountPointParser:
    def __init__(self) -> None:
        self._parser = _parser

    def parse_mount(self, expr: str) -> Mapping[str, str]:
        try:
            ast = self._parser.parse(expr)
            result = DictTransformer().transform(ast)
        except LarkError as e:
            raise ValueError(f"Virtual folder mount parsing error: {e}")
        return result



import asyncio
import logging
import time
from collections import defaultdict
from collections.abc import AsyncIterable
from typing import Any

import hiredis
from aiomonitor.task import preserve_termination_log
from aiotools.taskgroup import PersistentTaskGroup
from aiotools.taskgroup.types import AsyncExceptionHandler

from . import msgpack
from .events import AbstractEvent, EventHandler, _generate_consumer_id
from .events import EventDispatcher as _EventDispatcher
from .logging import BraceStyleAdapter
from .redis_client import RedisClient, RedisConnection
from .types import AgentId, EtcdRedisConfig

log = BraceStyleAdapter(logging.getLogger(__spec__.name))  # type: ignore[name-defined]

__all__ = ("EventDispatcher",)


async def read_stream(
    client: RedisClient,
    stream_key: str,
    *,
    block_timeout: int = 10_000,  # in msec
) -> AsyncIterable[tuple[bytes, dict[bytes, bytes]]]:
    """
    A high-level wrapper for the XREAD command.
    """
    last_id = b"$"
    while True:
        try:
            reply = await client.execute(
                ["XREAD", "BLOCK", block_timeout, "STREAMS", stream_key, last_id],
                command_timeout=(block_timeout + 5_000) / 1000,
            )
            if not reply:
                continue
            # Keep some latest messages so that other manager
            # processes to have chances of fetching them.
            await client.execute([
                "XTRIM",
                stream_key,
                "MAXLEN",
                "~",
                128,
            ])
            for msg_id, msg_data_list in reply[stream_key.encode()]:
                try:
                    msg_data = {}
                    for idx in range(0, len(msg_data_list), 2):
                        msg_data[msg_data_list[idx]] = msg_data_list[idx + 1]
                    yield msg_id, msg_data
                finally:
                    last_id = msg_id
        except asyncio.CancelledError:
            raise


async def read_stream_by_group(
    client: RedisClient,
    stream_key: str,
    group_name: str,
    consumer_id: str,
    *,
    autoclaim_idle_timeout: int = 1_000,  # in msec
    block_timeout: int = 10_000,  # in msec
) -> AsyncIterable[tuple[bytes, dict[bytes, bytes]]]:
    """
    A high-level wrapper for the XREADGROUP command
    combined with XAUTOCLAIM and XGROUP_CREATE.
    """
    while True:
        try:
            messages = []
            autoclaim_start_id = b"0-0"
            while True:
                reply = await client.execute(
                    [
                        "XAUTOCLAIM",
                        stream_key,
                        group_name,
                        consumer_id,
                        str(autoclaim_idle_timeout),
                        autoclaim_start_id,
                    ],
                    command_timeout=(autoclaim_idle_timeout + 5_000) / 1000,
                )
                if not reply:
                    continue

                for msg_id, msg_data_list in reply[1]:
                    msg_data = {}
                    for idx in range(0, len(msg_data_list), 2):
                        msg_data[msg_data_list[idx]] = msg_data_list[idx + 1]
                    messages.append((msg_id, msg_data))
                if reply[0] == b"0-0":
                    break
                autoclaim_start_id = reply[0]
            reply = await client.execute(
                [
                    "XREADGROUP",
                    "GROUP",
                    group_name,
                    consumer_id,
                    "BLOCK",
                    block_timeout,
                    "STREAMS",
                    stream_key,
                    ">",  # fetch messages not seen by other consumers
                ],
                command_timeout=(block_timeout + 5_000) / 1000,
            )
            if not reply:
                continue
            for msg_id, msg_data_list in reply[stream_key.encode()]:
                msg_data = {}
                for idx in range(0, len(msg_data_list), 2):
                    msg_data[msg_data_list[idx]] = msg_data_list[idx + 1]

                messages.append((msg_id, msg_data))
            await client.execute([
                "XACK",
                stream_key,
                group_name,
                *(msg_id for msg_id, msg_data in reply[stream_key.encode()]),
            ])
            for msg_id, msg_data in messages:
                yield msg_id, msg_data
        except asyncio.CancelledError:
            raise
        except hiredis.HiredisError as e:
            if e.args[0].startswith("NOGROUP "):
                try:
                    await client.execute([
                        "XGROUP",
                        "CREATE",
                        stream_key,
                        group_name,
                        "$",
                        "MKSTREAM",
                    ])
                except hiredis.HiredisError as e:
                    if e.args[0].startswith("BUSYGROUP "):
                        pass
                    else:
                        raise
                continue
            raise


class EventDispatcher(_EventDispatcher):
    redis_config: EtcdRedisConfig
    db: int
    consumers: defaultdict[str, set[EventHandler[Any, AbstractEvent]]]
    subscribers: defaultdict[str, set[EventHandler[Any, AbstractEvent]]]

    def __init__(
        self,
        redis_config: EtcdRedisConfig,
        db: int = 0,
        log_events: bool = False,
        *,
        consumer_group: str,
        service_name: str | None = None,
        stream_key: str = "events",
        node_id: str | None = None,
        consumer_exception_handler: AsyncExceptionHandler | None = None,
        subscriber_exception_handler: AsyncExceptionHandler | None = None,
    ) -> None:
        _redis_config = redis_config.copy()
        if service_name:
            _redis_config["service_name"] = service_name
        self.redis_config = redis_config.copy()
        self._log_events = True
        self.db = db
        self._closed = False
        self.consumers = defaultdict(set)
        self.subscribers = defaultdict(set)
        self._stream_key = stream_key
        self._consumer_group = consumer_group
        self._consumer_name = _generate_consumer_id(node_id)
        self.consumer_taskgroup = PersistentTaskGroup(
            name="consumer_taskgroup",
            exception_handler=consumer_exception_handler,
        )
        self.subscriber_taskgroup = PersistentTaskGroup(
            name="subscriber_taskgroup",
            exception_handler=subscriber_exception_handler,
        )

        self._log_events = log_events
        self.reconnect_poll_interval = 0.3

    def show_retry_warning(
        self,
        e: Exception,
        first_trial: float,
        retry_log_count: int,
        last_log_time: float,
        warn_on_first_attempt: bool = True,
    ) -> None:
        now = time.perf_counter()
        if (warn_on_first_attempt and retry_log_count == 0) or now - last_log_time >= 10.0:
            log.warning(
                "Retrying due to interruption of Redis connection " "({}, retrying-for: {:.3f}s)",
                repr(e),
                now - first_trial,
            )
            retry_log_count += 1
            last_log_time = now

    @preserve_termination_log
    async def _subscribe_loop(self) -> None:
        first_trial = time.perf_counter()
        retry_log_count = 0
        last_log_time = first_trial

        while True:
            try:
                async with RedisConnection(self.redis_config, db=self.db) as client:
                    async for msg_id, msg_data in read_stream(
                        client,
                        self._stream_key,
                    ):
                        if self._closed:
                            return
                        if msg_data is None:
                            continue
                        try:
                            await self.dispatch_subscribers(
                                msg_data[b"name"].decode(),
                                AgentId(msg_data[b"source"].decode()),
                                msgpack.unpackb(msg_data[b"args"]),
                            )
                        except asyncio.CancelledError:
                            raise
            except hiredis.HiredisError as e:
                if "READONLY" in e.args[0]:
                    self.show_retry_warning(e, first_trial, retry_log_count, last_log_time)
                    if self.reconnect_poll_interval:
                        await asyncio.sleep(self.reconnect_poll_interval)
                    continue
                elif "NOREPLICAS" in e.args[0]:
                    self.show_retry_warning(e, first_trial, retry_log_count, last_log_time)
                    if self.reconnect_poll_interval:
                        await asyncio.sleep(self.reconnect_poll_interval)
                    continue
                else:
                    raise
            except asyncio.TimeoutError as e:
                self.show_retry_warning(e, first_trial, retry_log_count, last_log_time)
                if self.reconnect_poll_interval:
                    await asyncio.sleep(self.reconnect_poll_interval)
                continue
            except (ConnectionError, EOFError) as e:
                self.show_retry_warning(e, first_trial, retry_log_count, last_log_time)
                if self.reconnect_poll_interval:
                    await asyncio.sleep(self.reconnect_poll_interval)
                continue
            except asyncio.CancelledError:
                raise
            except Exception:
                log.exception("EventDispatcher.subscribe(): unexpected-error")
                raise

    @preserve_termination_log
    async def _consume_loop(self) -> None:
        first_trial = time.perf_counter()
        retry_log_count = 0
        last_log_time = first_trial

        while True:
            try:
                async with RedisConnection(self.redis_config, db=self.db) as client:
                    async for msg_id, msg_data in read_stream_by_group(
                        client,
                        self._stream_key,
                        self._consumer_group,
                        self._consumer_name,
                    ):
                        if self._closed:
                            return
                        if msg_data is None:
                            continue
                        try:
                            await self.dispatch_consumers(
                                msg_data[b"name"].decode(),
                                AgentId(msg_data[b"source"].decode()),
                                msgpack.unpackb(msg_data[b"args"]),
                            )
                        except asyncio.CancelledError:
                            raise
            except hiredis.HiredisError as e:
                if "READONLY" in e.args[0]:
                    self.show_retry_warning(e, first_trial, retry_log_count, last_log_time)
                    if self.reconnect_poll_interval:
                        await asyncio.sleep(self.reconnect_poll_interval)
                    continue
                elif "NOREPLICAS" in e.args[0]:
                    self.show_retry_warning(e, first_trial, retry_log_count, last_log_time)
                    if self.reconnect_poll_interval:
                        await asyncio.sleep(self.reconnect_poll_interval)
                    continue
                else:
                    raise
            except asyncio.TimeoutError as e:
                self.show_retry_warning(e, first_trial, retry_log_count, last_log_time)
                if self.reconnect_poll_interval:
                    await asyncio.sleep(self.reconnect_poll_interval)
                continue
            except (ConnectionError, EOFError) as e:
                self.show_retry_warning(e, first_trial, retry_log_count, last_log_time)
                if self.reconnect_poll_interval:
                    await asyncio.sleep(self.reconnect_poll_interval)
                continue
            except asyncio.CancelledError:
                raise
            except Exception:
                log.exception("EventDispatcher.consume(): unexpected-error")
                raise

    async def close(self) -> None:
        self._closed = True



import datetime
import json
import uuid


class ExtendedJSONEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, uuid.UUID):
            return str(o)
        elif isinstance(o, datetime.datetime):
            return o.isoformat()
        return super().default(o)



# Relevant Linux kernel documentations
#
# For /proc filesystem, see
# https://docs.kernel.org/filesystems/proc.html
#
# For cgroup v1, see
# https://docs.kernel.org/admin-guide/cgroup-v1/
#
# For cgroup v2, see
# https://docs.kernel.org/admin-guide/cgroup-v2.html

import asyncio
import logging
from dataclasses import dataclass
from pathlib import Path
from typing import Literal, Optional

import aiohttp

from .docker import get_docker_connector
from .logging import BraceStyleAdapter
from .types import PID, ContainerId

log = BraceStyleAdapter(logging.getLogger(__spec__.name))  # type: ignore[name-defined]


@dataclass
class CgroupVersion:
    version: Literal["v1"] | Literal["v2"]
    driver: Literal["systemd"] | Literal["cgroupfs"]


async def get_docker_cgroup_version() -> CgroupVersion:
    connector = get_docker_connector()
    async with aiohttp.ClientSession(connector=connector.connector) as sess:
        async with sess.get(connector.docker_host / "info") as resp:
            data = await resp.json()
            return CgroupVersion(data["CgroupVersion"], data["CgroupDriver"])


def get_cgroup_mount_point(version: str, controller: str) -> Path:
    for line in Path("/proc/mounts").read_text().splitlines():
        device, mount_point, fstype, options, _ = line.split(" ", 4)
        match version:
            case "1":
                if fstype == "cgroup":
                    if controller in options.split(","):
                        return Path(mount_point)
            case "2":
                if fstype == "cgroup2":
                    return Path(mount_point)
    raise RuntimeError("could not find the cgroup mount point")


def get_cgroup_controller_id(controller: str) -> str:
    # example data
    # cpu <tab> 1 <tab> ...
    # cpuacct <tab> 1 <tab> ...
    for line in Path("/proc/cgroups").read_text().splitlines():
        name, id, _ = line.split("\t", 2)
        if name == controller:
            return id
    raise RuntimeError(f"could not find the cgroup controller {controller}")


def get_cgroup_of_pid(controller: str, pid: PID) -> str:
    # example data
    # 1:cpu,cpuacct:/<cgroup>
    controller_id = get_cgroup_controller_id(controller)
    for line in Path(f"/proc/{pid}/cgroup").read_text().splitlines():
        id, name, cgroup = line.split(":", 2)
        if id == controller_id:
            return cgroup.removeprefix("/")
    raise RuntimeError(f"could not find the cgroup of PID {pid}")


def get_container_id_of_cgroup(cgroup: str) -> Optional[str]:
    # cgroupfs driver: docker/<id>
    cgroupfs_prefix = "docker/"
    if cgroup.startswith(cgroupfs_prefix):
        return cgroup.removeprefix(cgroupfs_prefix)
    # systemd driver: system.slice/docker-<id>.scope
    systemd_prefix = "system.slice/docker-"
    systemd_suffix = ".scope"
    if cgroup.startswith(systemd_prefix) and cgroup.endswith(systemd_suffix):
        return cgroup.removeprefix(systemd_prefix).removesuffix(systemd_suffix)
    return None


async def get_container_pids(cid: ContainerId) -> list[int]:
    cgroup_version = await get_docker_cgroup_version()
    log.debug("Cgroup version: {}, {}", cgroup_version.version, cgroup_version.driver)
    match (cgroup_version.version, cgroup_version.driver):
        case ("2", "systemd"):
            tasks_path = Path(f"/sys/fs/cgroup/system.slice/docker-{cid}.scope/cgroup.procs")
        case ("2", "cgroupfs"):
            tasks_path = Path(f"/sys/fs/cgroup/docker/{cid}/cgroup.procs")
        case ("1", _):
            tasks_path = Path(f"/sys/fs/cgroup/pids/docker/{cid}/tasks")
        case _:
            raise RuntimeError("Should not reach here")
    tasks = await asyncio.get_running_loop().run_in_executor(None, tasks_path.read_text)
    return [*map(int, tasks.splitlines())]



import re
from typing import Iterator, List, Sequence, Set, Type

from .types import ServicePort, ServicePortProtocols

__all__ = ("parse_service_ports",)

_rx_service_ports = re.compile(
    r"^(?P<name>[\w-]+):(?P<proto>\w+):(?P<ports>\[\d+(?:,\d+)*\]|\d+)(?:,|$)"
)


def parse_service_ports(
    service_ports_label: str | Sequence[str],
    endpoint_ports_label: str | Sequence[str],
    exception_cls: Type[Exception] = ValueError,
) -> Sequence[ServicePort]:
    items: List[ServicePort] = []
    used_ports: Set[int] = set()
    inference_apps: Sequence[str]
    if isinstance(endpoint_ports_label, str):
        inference_apps = endpoint_ports_label.split(",")
    else:
        inference_apps = endpoint_ports_label

    def _iter_ports(s: str | Sequence[str]) -> Iterator[re.Match]:
        if isinstance(s, Sequence) and not isinstance(s, str):
            s = list(s)
            while s:
                piece = s.pop(0)
                match = _rx_service_ports.search(piece)
                if match:
                    yield match
        else:
            while True:
                match = _rx_service_ports.search(s)
                if match:
                    yield match
                    s = s[len(match.group(0)) :]
                else:
                    if len(s) > 0:
                        raise exception_cls("Invalid service-ports format")
                    break

    for match in _iter_ports(service_ports_label):
        name = match.group("name")
        if not name:
            raise exception_cls("Service port name must be not empty.")
        protocol = match.group("proto")
        if protocol == "pty":
            # unsupported, skip
            continue
        if protocol not in ("tcp", "http", "preopen"):
            raise exception_cls(f"Unsupported service port protocol: {protocol}")
        ports = tuple(map(int, match.group("ports").strip("[]").split(",")))
        for p in ports:
            if p in used_ports:
                raise exception_cls(f"The port {p} is already used by another service port.")
            if p <= 1024:
                raise exception_cls(
                    f"The service port number {p} must be "
                    "larger than 1024 to run without the root privilege."
                )
            if p >= 65535:
                raise exception_cls(f"The service port number {p} must be smaller than 65535.")
            if p in (2000, 2001, 2002, 2003, 2200, 7681):
                raise exception_cls(
                    "The service ports 2000 to 2003, 2200 and 7681 are reserved for internal use."
                )
            used_ports.add(p)
        items.append({
            "name": name,
            "protocol": ServicePortProtocols(protocol),
            "container_ports": ports,
            "host_ports": (None,) * len(ports),
            "is_inference": name in inference_apps,
        })

    return items
